#define _LINUX_

#ifndef _MINUBASEDDESC_H_
#define _MINUBASEDDESC_H_

//#define _SECURE_SCL 0

#ifndef _LINUX_
	#pragma warning (disable: 4996 4018 4244)
#endif

#pragma pack(1)

#include <vector>
#include <algorithm>
#define _USE_MATH_DEFINES
#include "math.h"
#include "assert.h"

#define _SHOW_WINDOW_

#ifdef _SHOW_WINDOW_
	#include "cv.h"
	#include "highgui.h"
#endif

#ifndef _SHOW_WINDOW_
	typedef unsigned char BYTE;
	typedef unsigned int DWORD;
	typedef int LONG;
	typedef unsigned short WORD;
#endif

#include "afis_live.h"
#include "MixArray.h"


using namespace std;


#ifdef _SHOW_WINDOW_
	#define _SHOW_INITIALPAIR_
	#define _SHOW_EACHMATCH_
	#define _SHOW_MISSMINU_
	#define _SHOW_BESTMATCH_
#endif

#ifndef MAX_PATH
#define MAX_PATH 260
#endif

// These two parameters need to be set by the caller
extern bool g_bPartial2Full;// true for partial-to-full match, false for full-to-full match
extern int g_nMaxRotation;// maximum rotation angle between two palmprints, 180 means that any rotation is possible

// parameters for MinutiaCode
const int MC_CNUM = 5;	// cirle number
const int MC_FNUM = 16;	// fan number
const int MC_LEN = MC_CNUM*MC_FNUM;
const int MC_R = 20;	// radius

typedef struct
{
	BYTE quality;
	BYTE group;
	BYTE dir;
	BYTE freq;
	BYTE nMinu[4];
}MinuCode;

// parameters for direction
const int DIR_INVALID = 91;

// parameters for frequency
const int FREQ_INVALID = 0;
const int FREQ_MIN = 3;
const int FREQ_MAX = 25;

extern bool g_bShowWindow; // When _SHOW_WINDOW_ is defined and g_bShowWindow is true, image is shown.

typedef struct
{
	int x;
	int y;
	int dir;//0-359
	int type;//1: endpoint; 3: bifurcation
	int quality;
}Minutia;

inline double MinuDist(Minutia m1, Minutia m2)
{
	return sqrt(((double)m1.x-m2.x)*(m1.x-m2.x)+(m1.y-m2.y)*(m1.y-m2.y));
}

class Point16
{
public:
	bool operator==(const Minutia& pt) {return (x==pt.x&&y==pt.y);}
	bool operator==(const Point16& pt) {return (x==pt.x&&y==pt.y);}
	Point16& operator=(const Minutia& pt) {x=pt.x;y=pt.y;return *this;}
	Point16& operator=(const Point16& pt) {x=pt.x;y=pt.y;return *this;}
	//Point16& operator-(const Point16& pt) {return Point16(x-pt.x,y-pt.y);}
	Point16() {x=0;y=0;}
	Point16(int x1, int y1) {x=x1;y=y1;}
	Point16(const Point16& pt) {x=pt.x;y=pt.y;}
	Point16(const Minutia& pt) {x=pt.x;y=pt.y;}
	double Dist(const Minutia& pt) { return sqrt(((double)x-pt.x)*(x-pt.x)+(y-pt.y)*(y-pt.y));}
	double Dist(const Point16& pt) { return sqrt(((double)x-pt.x)*(x-pt.x)+(y-pt.y)*(y-pt.y));}
	short int x,y;
};


typedef struct
{
	double dx,dy,rotS,rotC,differ;
	WORD xcenter,ycenter;
}Tran;

class Support
{
public:
	Support()
	{
		ii = -1; jj = -1; s = 0;
	}
	Support(short _i, short _j, double _s)
	{
		ii = _i; jj = _j; s = _s;
	}
	short ii;// index of minutia 
	short jj;// index of matched minutia of ii
	double s;// surport of the two minutiae
};

class MatchInfo
{
public:
	MatchInfo()
	{
		score = 0;
		pair_num = 0;
		desc_score = 0;
		minu_ratio1 = 0;
		minu_ratio2 = 0;
		relation_score = 0;
		align_error = 0;
		dirField_diff = 0;
		dirField_area = 0;
	}

	~MatchInfo()
	{
	}

	void Print()
	{
		printf("SCORE: %4.2f, NUM: %2d, DESC: %4.2f, RATIO: %4.2f, RELATION: %4.2f\n", score, pair_num, desc_score, minu_ratio1, relation_score);
	}

	static void WriteFeatureName(char* fname)
	{
		FILE* file = fopen(fname, "wb");
		fprintf(file, "%% 1-counter 2-f1 3-p1 4-f2 5-p2 ");
		fprintf(file, "6-score 7-pairNum 8-minuDesc 9-ratio1 10-ratio2 11-relation 12-alignError 13-dirDiff 14-dirArea\r\n");
		fclose(file);
	}

	void WriteFile(FILE* file);
	void WriteTransform(char* fname);
	void WriteMatchedMinuIndex(char* fname);
	void WriteMatchedMinuData(char* fname);
	bool NoPossibleMatch(Minutia minu, int k);
	void CompareDirField(Tran tran, int rot);
	void ComputeScore();

	// fingerprint information
	int		counter;			// 1
	int		f1, p1, f2, p2;		// 2-5

	// features
	double	score;				// the matching scores between images (all the following features are considered)
	int		pair_num;			// the number of matched minutiae
	double	desc_score;			// the similarity of descriptor of matched minutiae
	double	minu_ratio1;		// the ratio of the number of matched minutiae to the total number of minutiae in the common area of first images
	double	minu_ratio2;		// the ratio of the number of matched minutiae to the total number of minutiae in the common area of second images
	double	relation_score;		// a score measuring the similarity of the relationship between matched minutiae
	int		align_error;		// the average Euclidean distance between matched minutiae after alignment
	int		dirField_diff;		// the average difference of orientation fields of two images
	int		dirField_area;		// the number of blocks where two images have consistent ridge orientation

	vector<Support> mpairs;		// ID of matched minutiae
	Tran	m_tran;				// spatial transform between two images
};

#ifndef _SHOW_WINDOW_

	typedef struct tagBITMAPFILEHEADER {
	  WORD   bfType;
	  DWORD  bfSize;
	  WORD   bfReserved1;
	  WORD   bfReserved2;
	  DWORD  bfOffBits;
	} BITMAPFILEHEADER, *PBITMAPFILEHEADER;

	typedef struct tagBITMAPINFOHEADER{
	  DWORD  biSize;
	  LONG   biWidth;
	  LONG   biHeight;
	  WORD   biPlanes;
	  WORD   biBitCount;
	  DWORD  biCompression;
	  DWORD  biSizeImage;
	  LONG   biXPelsPerMeter;
	  LONG   biYPelsPerMeter;
	  DWORD  biClrUsed;
	  DWORD  biClrImportant;
	} BITMAPINFOHEADER, *PBITMAPINFOHEADER;
#endif

typedef vector<Point16> Ridge;
typedef vector<int> VecInt;
typedef vector<VecInt> Vec2Int;
typedef vector<Vec2Int> Vec3Int;
typedef vector<BYTE> VecByte;
typedef vector<VecByte> Vec2Byte;
typedef vector<bool> VecBool;
typedef vector<VecBool> VecBools;
typedef vector<double> VecDb;
typedef vector<VecDb> VecDbs;
typedef vector<Support> SupportList;


//////////////////////////////////////////////////////////////////////////
// (-180,180]
inline int NormalizeAngle(int a)
{
	if(a>=360)
		a-=360;
	else if(a<=-360)
		a+=360;
	else if(a>180)
		a-=360;
	else if(a<-180)
		a+=360;
	return a;
}

inline int NormalizeAngle180(int a)
{
	int b = a;
	if(b>=90)
		b -= 180;
	else if(b<=-90)
		b += 180;
	return b;
}

#define min(x,y) (((x)>(y))?(y):(x))

#endif
